import os
import sys
import threading
import subprocess

from ._perpetuo import StallTracker


WATCHER: subprocess.Popen | None = None
GIL_STALLTRACKER: StallTracker | None = None


def start_watcher(
    *,
    poll_interval: float | None = None,
    alert_interval: float | None = None,
    traceback_suppress: float | None = None,
    print_locals: bool = True,
) -> None:
    global WATCHER
    if WATCHER is None:
        args = []
        if poll_interval is not None:
            args += ["--poll-interval", str(poll_interval)]
        if alert_interval is not None:
            args += ["--alert-interval", str(alert_interval)]
        if traceback_suppress is not None:
            args += ["--traceback-suppress", str(alert_interval)]
        if print_locals:
            args += ["--print-locals"]
        else:
            args += ["--no-print-locals"]
        subprocess.Popen(["perpetuo", *args, "watch", str(os.getpid())])


def stop_watcher() -> None:
    global WATCHER
    if WATCHER is not None:
        WATCHER.kill()
        WATCHER.wait()
        WATCHER = None


def instrument_gil() -> None:
    global GIL_STALLTRACKER
    if GIL_STALLTRACKER is not None:
        return
    if hasattr(sys, "_set_stall_counter"):
        GIL_STALLTRACKER = StallTracker("GIL", "gil")
        # Conceptually, sys._set_stall_counter holds a reference to this object, so we
        # do an intentionally unbalanced incref here. In particular, this avoids the
        # StallTracker getting GC'ed when the interpreter shuts down and clears all
        # module globals, while it's actually still in use.
        GIL_STALLTRACKER._leak()
        sys._set_stall_counter(GIL_STALLTRACKER.counter_address())
    else:
        raise RuntimeError(
            "This Python was not built with the perpetuo GIL instrumentation patch"
        )


def instrument_trio() -> None:
    import trio

    class TrioStallInstrument(trio.abc.Instrument):
        def __init__(self):
            # We construct the tracker lazily, to make sure that we're already in Trio
            # before it becomes active
            self.stall_tracker: StallTracker | None = None

        def _init(self):
            assert self.stall_tracker is None
            thread = threading.current_thread().ident
            self.stall_tracker = StallTracker(
                f"Trio run loop (thread {thread:#_x})", thread
            )

        def before_task_step(self, _: trio.lowlevel.Task) -> None:
            if self.stall_tracker is None:
                self._init()
                # New StallTracker starts out in the "active" state, which is what we
                # want
            else:
                self.stall_tracker.go_active()

        def after_task_step(self, _: trio.lowlevel.Task) -> None:
            if self.stall_tracker is None:
                self._init()
                assert self.stall_tracker is not None
                # New StallTracker starts out in the "active" state, so need to toggle
                # it
            self.stall_tracker.go_idle()

        def after_run(self) -> None:
            self.stall_tracker.close()

    trio.lowlevel.add_instrument(TrioStallInstrument())


def dwim(**kwargs) -> list[str]:
    did = []

    try:
        instrument_gil()
    except RuntimeError:
        pass
    else:
        did.append("instrumented GIL")

    if "trio" in sys.modules:
        try:
            instrument_trio()
        except RuntimeError:
            pass
        else:
            did.append("instrumented Trio")

    if did:
        start_watcher(**kwargs)
        did.append("started out of process watcher")

    return did



import pytest
import perpetuo
import trio
import time


def test_repeated_instantiation():
    # On x86-64, our export page is 4096 bytes, and each StallTacker has multiple
    # pointers, so >16 bytes each. So 1000 would definitely be enough to overflow, if we
    # aren't freeing them.
    for _ in range(1000):
        st = perpetuo.StallTracker("test", 1)
        st.go_idle()
        st.close()


# Doesn't really test much without the instrumentation patch, but even on vanilla python
# I guess it's good to check it doesn't crash
def test_repeated_gil():
    for _ in range(1000):
        try:
            perpetuo.instrument_gil()
        except RuntimeError:
            pass


def test_repeated_trio():
    for _ in range(1000):

        async def main():
            perpetuo.instrument_trio()
            await trio.lowlevel.checkpoint()

        trio.run(main)


# XX FIXME: figure out how to run this test in CI without breaking everywhere else,
# given the permission problems on macOS/Linux
@pytest.mark.skip
def test_catches_stall(capfd):
    assert "stall detected in process" not in capfd.readouterr()
    try:

        async def main():
            print(perpetuo.dwim())
            time.sleep(1)

        trio.run(main)

        assert "stall detected in process" in capfd.readouterr()
    finally:
        perpetuo.stop_watcher()



from typing import Literal

class StallTracker:
    def __init__(self, name: str, relevant_thread: int | Literal["gil"]): ...
    def go_active(self) -> None: ...
    def go_idle(self) -> None: ...
    def is_active(self) -> bool: ...
    def counter_address(self) -> int: ...
    def close(self) -> None: ...
    def _leak(self) -> None: ...

def stall_gil(seconds: float) -> None: ...



from setuptools import setup, find_packages
from setuptools_rust import Binding, RustBin, RustExtension, Strip
from wheel.bdist_wheel import bdist_wheel


# https://github.com/joerick/python-abi3-package-sample/blob/main/setup.py
class bdist_wheel_abi3(bdist_wheel):
    def get_tag(self):
        python, abi, plat = super().get_tag()

        if python.startswith("cp"):
            # on CPython, our wheels are abi3 and compatible back to 3.9
            return "cp39", "abi3", plat

        return python, abi, plat


setup(
    name="perpetuo",
    rust_extensions=[
        RustExtension(
            "perpetuo._perpetuo",
            binding=Binding.PyO3,
            py_limited_api=True,
            # strip=Strip.Debug
        ),
        RustBin("perpetuo", strip=Strip.Debug),
    ],
    cmdclass={"bdist_wheel": bdist_wheel_abi3},
    package_dir={"": "python"},
    packages=find_packages(where="python"),
    package_data={"perpetuo": ["py.typed", "__init__.pyi"]},
    zip_safe=False,
)



__all__ = [
    "StallTracker",
    "stall_gil",
    "start_watcher",
    "stop_watcher",
    "instrument_gil",
    "instrument_trio",
    "dwim",
]

from ._perpetuo import StallTracker, stall_gil
from ._setup import start_watcher, stop_watcher, instrument_gil, instrument_trio, dwim
